(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('typen')) :
  typeof define === 'function' && define.amd ? define(['exports', 'typen'], factory) :
  (global = global || self, factory(global.borel = {}, global.typen));
}(this, (function (exports, typen) { 'use strict';

  // import { quickSort, quickSortBy } from './sort/quickSort'
  // import { timSort } from './sort/timSort'
  // If compareFunction(a, b) is less than 0, sort a to an index lower than b (i.e. a comes first).
  // If compareFunction(a, b) returns 0, leave a and b unchanged with respect to each other.
  // If compareFunction(a, b) is greater than 0, sort b to an index lower than a (i.e. b comes first).
  var Comparer =
  /*#__PURE__*/
  function () {
    function Comparer() {}

    Comparer.stringAscending = function stringAscending(a, b) {
      return a.localeCompare(b);
    };

    Comparer.stringDescending = function stringDescending(a, b) {
      return b.localeCompare(a);
    };

    Comparer.numberAscending = function numberAscending(a, b) {
      return a - b;
    };

    Comparer.numberDescending = function numberDescending(a, b) {
      return b - a;
    };

    return Comparer;
  }();

  var Rank =
  /*#__PURE__*/
  function () {
    function Rank() {}

    /**
     *
     * @param {*[]} arr
     * @param {function(*,*):number} comparer Compare 'prev' & 'next' element in an array. If return negative number or
     * zero, then 'prev' comes first, else 'next' comes first.
     * @return {*[]} Sorted arr.
     */
    Rank.sort = function sort(arr, comparer) {
      return arr.slice().sort(comparer);
    }
    /**
     *
     * @param {*[]} arr
     * @param {function(*,*):number} comparer Compare 'prev' & 'next' element in an array. If return < 0, 'prev' comes first. If return > 0, 'next' comes first.
     * @param {function(*):boolean} [filter]
     * @return {number[]} Rank order array, where 0 denote the first.
     */
    ;

    Rank.rank = function rank(arr, comparer, filter) {
      var sorted = (!filter ? arr.slice() : arr.filter(filter)).sort(comparer);
      return arr.map(function (x) {
        return sorted.indexOf(x);
      });
    }
    /**
     *
     * @param {*[]} arr
     * @param {number[]} arrRank array of the same length as 'arr', containing rank order of 'arr', 0 comes first.
     * @return {*[]}
     */
    ;

    Rank.reorderBy = function reorderBy(arr, arrRank) {
      var rsl = Array(arr.length);

      for (var _iterator = arrRank.entries(), _isArray = Array.isArray(_iterator), _i = 0, _iterator = _isArray ? _iterator : _iterator[Symbol.iterator]();;) {
        var _ref;

        if (_isArray) {
          if (_i >= _iterator.length) break;
          _ref = _iterator[_i++];
        } else {
          _i = _iterator.next();
          if (_i.done) break;
          _ref = _i.value;
        }

        var _ref2 = _ref,
            i = _ref2[0],
            ord = _ref2[1];
        rsl[ord] = arr[i];
      }

      return rsl;
    };

    return Rank;
  }();

  var isNumLcher = function isNumLcher(level) {
    if (level === void 0) {
      level = 0;
    }

    switch (level) {
      case 0:
        return function (x) {
          return !isNaN(x);
        };

      case 1:
        return typen.NumLoose.isNumeric;

      case 2:
      default:
        return typen.Num.isNumeric;
    }
  };
  var toNumLcher = function toNumLcher(level) {
    if (level === void 0) {
      level = 0;
    }

    switch (level) {
      case 0:
        return function (x) {
          return x;
        };

      case 1:
        return typen.NumLoose.numeric;

      case 2:
      default:
        return typen.Num.numeric;
    }
  };

  var locNumInAr = function locNumInAr(arr, sta, end, _temp) {
    var _ref = _temp === void 0 ? {} : _temp,
        _ref$level = _ref.level,
        level = _ref$level === void 0 ? 0 : _ref$level;

    for (var el, isNum = isNumLcher(level); sta < end; sta++) {
      el = arr[sta];
      if (isNum(el)) return [sta, el];
    }

    return [end, NaN];
  };
  var locNumInMx = function locNumInMx(mx, ri, re, ci, ce, _temp2) {
    var _ref2 = _temp2 === void 0 ? {} : _temp2,
        _ref2$level = _ref2.level,
        level = _ref2$level === void 0 ? 0 : _ref2$level;

    for (var el, isNum = isNumLcher(level); ri < re; ri++) {
      for (ci = 0; ci < ce; ci++) {
        el = mx[ri][ci];
        if (isNum(el)) return [ri, ci, el];
      }
    }

    return [re, ce, NaN];
  };
  var locNumInCol = function locNumInCol(mx, ri, re, c, _temp3) {
    var _ref3 = _temp3 === void 0 ? {} : _temp3,
        _ref3$level = _ref3.level,
        level = _ref3$level === void 0 ? 0 : _ref3$level;

    for (var el, isNum = isNumLcher(level); ri < re; ri++) {
      el = mx[ri][c];
      if (isNum(el)) return [ri, el];
    }

    return [re, NaN];
  };

  var toBound = function toBound(max, min, dif) {
    return dif ? {
      min: min,
      dif: max - min
    } : {
      max: max,
      min: min
    };
  };

  function sortEntries(entries, sort) {
    switch (sort) {
      case true:
      case 'desc':
        return entries.sort(function (_ref, _ref2) {
          var a = _ref[1];
          var b = _ref2[1];
          return b - a;
        });

      case 'asc':
        return entries.sort(function (_ref3, _ref4) {
          var a = _ref3[1];
          var b = _ref4[1];
          return a - b;
        });

      case false:
      default:
        return entries;
    }
  }

  var plusOrZero = function plusOrZero(x) {
    return x ? x + 1 : 1;
  };

  var Stat =
  /*#__PURE__*/
  function () {
    function Stat() {}

    Stat.cnt = function cnt(arr) {
      return !!arr ? arr.length : 0;
    }
    /**
     *
     * @param {number[]} arr
     * @param {number} [level] - 0: no check, 1: loose check, 2: strict check
     * @returns {number}
     */
    ;

    Stat.sum = function sum(arr, _temp) {
      var _ref = _temp === void 0 ? {} : _temp,
          _ref$level = _ref.level,
          level = _ref$level === void 0 ? 0 : _ref$level;

      if (!arr) return 0;
      var tr = toNumLcher(level);
      var l = arr.length;

      switch (l) {
        case 0:
          return NaN;

        case 1:
          return level ? tr(arr[0]) : arr[0];

        default:
          var sum = 0;

          if (level) {
            for (--l; l >= 0; l--) {
              sum += tr(arr[l]);
            }
          } else {
            for (--l; l >= 0; l--) {
              sum += arr[l];
            }
          }

          return sum;
      }
    }
    /**
     *
     * @param {number[]} arr
     * @param {number} [level] - 0: no check, 1: loose check, 2: strict check
     * @returns {number}
     */
    ;

    Stat.avg = function avg(arr, _temp2) {
      var _ref2 = _temp2 === void 0 ? {} : _temp2,
          _ref2$level = _ref2.level,
          level = _ref2$level === void 0 ? 0 : _ref2$level;

      return arr && arr.length ? Stat.sum(arr, {
        level: level
      }) / arr.length : 0;
    };

    Stat.mode = function mode(arr) {
      return undefined;
    }
    /**
     *
     * @param {*[]} arr
     * @param {boolean} [count=false]
     * @param {string|boolean} [sort=false]
     * @returns {[any, any][]|[]|any[]|*}
     */
    ;

    Stat.distinct = function distinct(arr, _temp3) {
      var _ref3 = _temp3 === void 0 ? {} : _temp3,
          _ref3$count = _ref3.count,
          count = _ref3$count === void 0 ? false : _ref3$count,
          _ref3$sort = _ref3.sort,
          sort = _ref3$sort === void 0 ? false : _ref3$sort;

      if (!arr) return count ? [[]] : [];
      var l = arr.length;
      if (!l) return count ? [[]] : [];

      if (count) {
        var ents;

        if (l <= 0x100) {
          ents = [];

          var _loop = function _loop(i, _j, _el) {
            _el = arr[i];
            _j = ents.findIndex(function (x) {
              return _el === x[0];
            });
            if (_j < 0) _j += ents.push([_el, 0]);
            ents[_j][1]++;
            j = _j;
            el = _el;
          };

          for (var i = 0, j, el; i < l; i++) {
            _loop(i, j, el);
          }
        } else {
          var lx = new Map();

          for (var _i = 0, _el2; _i < l; _i++) {
            _el2 = arr[_i];
            lx.set(_el2, plusOrZero(lx.get(_el2)));
          }

          ents = Array.from(lx.entries());
        }

        return sortEntries(ents, sort);
      } else {
        if (l <= 0x100) {
          var dist = [];

          for (var _i2 = 0; _i2 < l; _i2++) {
            if (dist.indexOf(arr[_i2]) < 0) dist.push(arr[_i2]);
          }

          return dist;
        } else {
          return Array.from(new Set(arr));
        }
      }
    };

    Stat.median = function median(arr) {
      return undefined;
    }
    /**
     *
     * @param {*[]} arr
     * @param {boolean} [dif=false]
     * @param {number} [level=0]
     * @returns {{min: *, max: *}|{min: *, dif: *}}}
     */
    ;

    Stat.bound = function bound(arr, _temp4) {
      var _ref4 = _temp4 === void 0 ? {} : _temp4,
          _ref4$dif = _ref4.dif,
          dif = _ref4$dif === void 0 ? false : _ref4$dif,
          _ref4$level = _ref4.level,
          level = _ref4$level === void 0 ? 0 : _ref4$level;

      var l = arr === null || arr === void 0 ? void 0 : arr.length;
      if (!l) return toBound(NaN, NaN, dif);
      var t = toNumLcher(level);

      var _locNumInAr = locNumInAr(arr, 0, l, {
        level: level
      }),
          i = _locNumInAr[0],
          el = _locNumInAr[1];

      var max = t(el),
          min = max;

      for (--l; l > i; --l) {
        el = t(arr[l]);

        if (el < min) {
          min = el;
        } else if (el > max) {
          max = el;
        }
      }

      return toBound(max, min, dif);
    } // Population standard deviation
    ;

    Stat.stDevP = function stDevP(arr, _temp5) {
      var _ref5 = _temp5 === void 0 ? {} : _temp5,
          _ref5$level = _ref5.level,
          level = _ref5$level === void 0 ? 0 : _ref5$level;

      var cnt = Stat.cnt(arr);
      if (!cnt) return 0;
      var avg = Stat.avg(arr, {
        level: level
      });
      return Math.hypot.apply(Math, arr.map(function (x) {
        return x - avg;
      })) / Math.sqrt(cnt);
    } // Sample standard deviation
    ;

    Stat.stDevS = function stDevS(arr, _temp6) {
      var _ref6 = _temp6 === void 0 ? {} : _temp6,
          _ref6$level = _ref6.level,
          level = _ref6$level === void 0 ? 0 : _ref6$level;

      var cnt = Stat.cnt(arr) - 1;
      if (!cnt) return 0;
      var avg = Stat.avg(arr, {
        level: level
      });
      return Math.hypot.apply(Math, arr.map(function (x) {
        return x - avg;
      })) / Math.sqrt(cnt);
    }
    /**
     *
     * @param {*[]} arr
     * @param {function(*):number} ject
     * @return {number}
     */
    ;

    Stat.sumBy = function sumBy(arr, ject) {
      if (!arr) return NaN;
      var l = arr.length;

      switch (l) {
        case 0:
          return NaN;

        case 1:
          return ject(arr[0]);

        default:
          var sum = 0;

          for (var i = 0; i < l; i++) {
            sum += ject(arr[i]);
          }

          return sum;
      }
    }
    /**
     *
     * @param {*[]} arr
     * @param {function(*):number} ject
     * @return {number}
     */
    ;

    Stat.maxBy = function maxBy(arr, ject) {
      if (!arr) return NaN;
      var l = arr.length;

      switch (l) {
        case 0:
          return NaN;

        case 1:
          return ject(arr[0]);

        default:
          var v,
              max = ject(arr[0]);

          for (var i = 1; i < l; i++) {
            v = ject(arr[i]);
            if (v > max) max = v;
          }

          return max;
      }
    };

    return Stat;
  }();
  //   let rsl = 0;
  //   for (let i = 0; i < arr.length; i++) {
  //     rsl += arr[i];
  //   }
  //   return rsl;
  // }
  // function sum (arr) {
  //   switch (arr.length) {
  //     case 0:
  //       return NaN
  //     case 1:
  //       return arr[0]
  //     default:
  //       let sum = 0
  //       for (let n of arr) {
  //         sum += n
  //       }
  //       return sum
  //   }
  // }

  var size = function size(mx) {
    var _mx$;

    var l = mx === null || mx === void 0 ? void 0 : mx.length;
    return [l, l ? (_mx$ = mx[0]) === null || _mx$ === void 0 ? void 0 : _mx$.length : undefined];
  };

  var StatMx =
  /*#__PURE__*/
  function () {
    function StatMx() {}

    StatMx.cnt = function cnt(mx) {
      if (mx && mx.length) {
        var row = mx[0];
        if (row && row.length) return length * row.length;
      }

      return 0;
    } // /**
    //  *
    //  * @param {number[] }mx
    //  * @returns {number}
    //  */
    // static sum (mx) {
    //   if (!mx) return 0
    //   let { length: l } = mx
    //   switch (l) {
    //     case 0:
    //       return NaN
    //     case 1:
    //       return mx[0]
    //     default:
    //       let sum = 0
    //       for (--l; l >= 0; l--) sum += mx[l]
    //       return sum
    //   }
    // }
    //
    // static avg (mx) {
    //   const cnt = StatMx.cnt(mx)
    //   return cnt ? StatMx.sum(mx) / cnt : 0
    // }

    /**
     *
     * @param {*[]} mx
     * @param {boolean} [dif=false]
     * @param {number} [level=0]
     * @returns {{min: *, max: *}|{min: *, dif: *}}}
     */
    ;

    StatMx.bound = function bound(mx, _temp) {
      var _ref = _temp === void 0 ? {} : _temp,
          _ref$dif = _ref.dif,
          dif = _ref$dif === void 0 ? false : _ref$dif,
          _ref$level = _ref.level,
          level = _ref$level === void 0 ? 0 : _ref$level;

      var _size = size(mx),
          h = _size[0],
          w = _size[1];

      if (!h || !w) return toBound(NaN, NaN, dif);
      var t = toNumLcher(level);

      var _locNumInMx = locNumInMx(mx, 0, h, 0, w, {
        level: level
      }),
          i = _locNumInMx[0],
          el = _locNumInMx[2],
          max = t(el),
          min = max,
          _a,
          _i;

      for (--h; h > i; --h) {
        var _Stat$bound = Stat.bound(mx[h], {
          level: level
        });

        _a = _Stat$bound.max;
        _i = _Stat$bound.min;

        if (_i < min) {
          min = _i;
        } else if (_a > max) {
          max = _a;
        }
      }

      return toBound(max, min, dif);
    }
    /**
     *
     * @param {*[]} mx
     * @param {number} y
     * @param {boolean} [dif=false]
     * @param {number} [level=0] - level: 0, none; 1, loose; 2, strict
     * @returns {{min: *, max: *}|{min: *, dif: *}}}
     */
    ;

    StatMx.boundCol = function boundCol(mx, y, _temp2) {
      var _ref2 = _temp2 === void 0 ? {} : _temp2,
          _ref2$dif = _ref2.dif,
          dif = _ref2$dif === void 0 ? false : _ref2$dif,
          _ref2$level = _ref2.level,
          level = _ref2$level === void 0 ? 0 : _ref2$level;

      var _size2 = size(mx),
          h = _size2[0],
          w = _size2[1];

      if (!h || !w || y >= w) return toBound(NaN, NaN, dif);
      var t = toNumLcher(level);

      var _locNumInCol = locNumInCol(mx, 0, h, y, {
        level: level
      }),
          i = _locNumInCol[0],
          el = _locNumInCol[1],
          max = t(el),
          min = max;

      for (--h; h > i; --h) {
        el = t(mx[h][y]);

        if (el < min) {
          min = el;
        } else if (el > max) {
          max = el;
        }
      }

      return toBound(max, min, dif);
    };

    StatMx.distinct = function distinct(mx, y) {
      var _size3 = size(mx),
          h = _size3[0],
          w = _size3[1];

      if (!h || !w || y >= w) return [[]];
      var rows = [];

      var _loop = function _loop(i, _r2) {
        _r2 = mx[i];
        if (rows.findIndex(function (r) {
          return _r2[y] === r[y];
        }) < 0) rows.push(_r2);
        _r = _r2;
      };

      for (var i = 0, _r; i < h; i++) {
        _loop(i, _r);
      }

      return rows;
    }
    /**
     *
     * @param {*[]} mx
     * @param {number} y
     * @param {boolean} [count=false]
     * @param {string|boolean} [sort=false] - When sort is function, sort must be a comparer between two point element.
     * @returns {[any, any][]|[]|any[]|*}
     */
    ;

    StatMx.distinctCol = function distinctCol(mx, y, _temp3) {
      var _ref3 = _temp3 === void 0 ? {} : _temp3,
          _ref3$count = _ref3.count,
          count = _ref3$count === void 0 ? false : _ref3$count,
          _ref3$sort = _ref3.sort,
          sort = _ref3$sort === void 0 ? false : _ref3$sort;

      var _size4 = size(mx),
          h = _size4[0],
          w = _size4[1];

      if (!h || !w || y >= w) return count ? [[]] : [];

      if (!count) {
        var dist = [];

        for (var i = 0, el; i < h; i++) {
          el = mx[i][y];
          if (dist.indexOf(el) < 0) dist.push(el);
        }

        return dist;
      } else {
        var ents = [];

        var _loop2 = function _loop2(_j, _i2, _el2) {
          _el2 = mx[_i2][y];
          _j = ents.findIndex(function (x) {
            return _el2 === x[0];
          });
          if (_j < 0) _j += ents.push([_el2, 0]);
          ents[_j][1]++;
          j = _j;
          _el = _el2;
        };

        for (var _i2 = 0, j, _el; _i2 < h; _i2++) {
          _loop2(j, _i2, _el);
        }

        return sortEntries(ents, sort);
      }
    };

    return StatMx;
  }();

  var Zu =
  /*#__PURE__*/
  function () {
    function Zu() {}

    /**
     * Generate a random integer between [min, max].
     * Both min & max are inclusive.
     * @param {Number} min  Int
     * @param {Number} max  Int
     * @returns {Number}  Int
     */
    Zu.randBetween = function randBetween(min, max) {
      return ~~(Math.random() * (max - min + 1)) + min;
    }
    /**
     * Generate a random integer between [min, max).
     * Notice: min is inclusive & max is exclusive.
     * @param {Number} min  Int
     * @param {Number} max(exclusive)  Int
     * @returns {Number}  Int
     */
    ;

    Zu.rand = function rand(min, max) {
      return ~~(Math.random() * (max - min)) + min;
    };

    Zu.almostEquals = function almostEquals(x, y, epsilon) {
      return Math.abs(x - y) < epsilon;
    };

    Zu.almostInt = function almostInt(x, epsilon) {
      // let rounded = Math.round(x)
      // return rounded - epsilon < x && rounded + epsilon > x
      return Math.abs(x - Math.round(x)) < epsilon;
    };

    Zu.intExponent = function intExponent(x) {
      return ~~Math.log10(x);
    };

    return Zu;
  }();
  //   ? x => Math.log10(x)
  //   : x => {
  //     let exponent = Math.log(x) * Math.LOG10E // Math.LOG10E = 1 / Math.LN10.
  //     // Check for whole powers of 10,
  //     // which due to floating point rounding error should be corrected.
  //     let powerOf10 = Math.round(exponent)
  //     let isPowerOf10 = x === Math.pow(10, powerOf10)
  //     return isPowerOf10 ? powerOf10 : exponent
  //   }
  // /**
  //  * Generate a random integer between [min, max].
  //  * Both min & max are inclusive.
  //  * @param {Number} min  Int
  //  * @param {Number} max  Int
  //  * @returns {Number}  Int
  //  */
  // function randBetween (min, max) {
  //   return Math.floor(Math.random() * (max - min + 1)) + min
  // }

  exports.Comparer = Comparer;
  exports.Rank = Rank;
  exports.Stat = Stat;
  exports.StatMx = StatMx;
  exports.Zu = Zu;

  Object.defineProperty(exports, '__esModule', { value: true });

})));
